@@ -4,6 +4,7 @@ use crate::ast;
 use crate::error::Error;
 use crate::instructions::{Instruction, Instructions};
 use crate::tokens::Span;
+use crate::utils::matches;
 use crate::value::Value;
 
 /// Represents an open block of code that does not yet have updated
@@ -282,6 +283,16 @@ impl<'source> Compiler<'source> {
                 }
                 self.add(Instruction::PopAutoEscape);
             }
+            ast::Stmt::FilterBlock(filter_block) => {
+                self.set_location_from_span(filter_block.span());
+                self.add(Instruction::BeginCapture);
+                for node in &filter_block.body {
+                    self.compile_stmt(node)?;
+                }
+                self.add(Instruction::EndCapture);
+                self.compile_expr(&filter_block.filter)?;
+                self.add(Instruction::Emit);
+            }
         }
         Ok(())
     }
@@ -371,7 +382,9 @@ impl<'source> Compiler<'source> {
             }
             ast::Expr::Filter(f) => {
                 self.set_location_from_span(f.span());
-                self.compile_expr(&f.expr)?;
+                if let Some(ref expr) = f.expr {
+                    self.compile_expr(expr)?;
+                }
                 for arg in &f.args {
                     self.compile_expr(arg)?;
                 }
@@ -408,6 +421,11 @@ impl<'source> Compiler<'source> {
                         self.add(Instruction::BuildList(c.args.len()));
                         self.add(Instruction::CallFunction(name));
                     }
+                    ast::CallType::Block(name) => {
+                        self.add(Instruction::BeginCapture);
+                        self.add(Instruction::CallBlock(name));
+                        self.add(Instruction::EndCapture);
+                    }
                     ast::CallType::Method(expr, name) => {
                         self.compile_expr(expr)?;
                         for arg in &c.args {
