@@ -8,6 +8,7 @@ pub struct Error {
     detail: Option<Cow<'static, str>>,
     name: Option<String>,
     lineno: usize,
+    source: Option<Box<dyn std::error::Error>>,
 }
 
 impl PartialEq for Error {
@@ -25,6 +26,7 @@ pub enum ErrorKind {
     NonPrimitive,
     NonKey,
     ImpossibleOperation,
+    InvalidOperation,
     SyntaxError,
     TemplateNotFound,
     InvalidArguments,
@@ -32,6 +34,7 @@ pub enum ErrorKind {
     UnknownTest,
     BadEscape,
     UndefinedError,
+    BadSerialization,
 }
 
 impl ErrorKind {
@@ -41,6 +44,7 @@ impl ErrorKind {
             ErrorKind::NonPrimitive => "not a primitive",
             ErrorKind::NonKey => "not a key type",
             ErrorKind::ImpossibleOperation => "impossible operation",
+            ErrorKind::InvalidOperation => "invalid operation",
             ErrorKind::SyntaxError => "syntax error",
             ErrorKind::TemplateNotFound => "template not found",
             ErrorKind::InvalidArguments => "invalid arguments",
@@ -48,6 +52,7 @@ impl ErrorKind {
             ErrorKind::UnknownTest => "unknown test",
             ErrorKind::BadEscape => "bad string escape",
             ErrorKind::UndefinedError => "variable or attribute undefined",
+            ErrorKind::BadSerialization => "could not serialize to internal format",
         }
     }
 }
@@ -80,6 +85,7 @@ impl Error {
             detail: Some(detail.into()),
             name: None,
             lineno: 0,
+            source: None,
         }
     }
 
@@ -88,6 +94,12 @@ impl Error {
         self.lineno = lineno;
     }
 
+    #[allow(unused)]
+    pub(crate) fn with_source<E: std::error::Error + 'static>(mut self, source: E) -> Self {
+        self.source = Some(Box::new(source));
+        self
+    }
+
     /// Returns the error kind
     pub fn kind(&self) -> ErrorKind {
         self.kind
@@ -104,7 +116,11 @@ impl Error {
     }
 }
 
-impl std::error::Error for Error {}
+impl std::error::Error for Error {
+    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
+        self.source.as_deref()
+    }
+}
 
 impl From<ErrorKind> for Error {
     fn from(kind: ErrorKind) -> Self {
@@ -113,6 +129,16 @@ impl From<ErrorKind> for Error {
             detail: None,
             name: None,
             lineno: 0,
+            source: None,
         }
     }
 }
+
+impl serde::ser::Error for Error {
+    fn custom<T>(msg: T) -> Self
+    where
+        T: fmt::Display,
+    {
+        Error::new(ErrorKind::BadSerialization, msg.to_string())
+    }
+}
