@@ -2,9 +2,12 @@ use crate::ast::{self, Spanned};
 use crate::error::{Error, ErrorKind};
 use crate::lexer::tokenize;
 use crate::tokens::{Span, Token};
+use crate::utils::matches;
 use crate::value::Value;
 
-const RESERVED_NAMES: [&str; 7] = ["true", "True", "false", "False", "none", "None", "loop"];
+const RESERVED_NAMES: [&str; 8] = [
+    "true", "True", "false", "False", "none", "None", "loop", "self",
+];
 
 macro_rules! syntax_error {
     ($msg:expr) => {{
@@ -105,7 +108,6 @@ impl<'a> TokenStream<'a> {
 }
 
 struct Parser<'a> {
-    filename: &'a str,
     stream: TokenStream<'a>,
 }
 
@@ -156,9 +158,8 @@ macro_rules! unaryop {
 }
 
 impl<'a> Parser<'a> {
-    pub fn new(source: &'a str, filename: &'a str, in_expr: bool) -> Parser<'a> {
+    pub fn new(source: &'a str, in_expr: bool) -> Parser<'a> {
         Parser {
-            filename,
             stream: TokenStream::new(source, in_expr),
         }
     }
@@ -325,17 +326,23 @@ impl<'a> Parser<'a> {
                         Vec::new()
                     };
                     expr = ast::Expr::Filter(Spanned::new(
-                        ast::Filter { name, expr, args },
+                        ast::Filter {
+                            name,
+                            expr: Some(expr),
+                            args,
+                        },
                         self.stream.expand_span(span),
                     ));
                 }
                 Some((Token::Ident("is"), _)) => {
                     self.stream.next()?;
-                    let mut negated = false;
-                    if matches!(self.stream.current()?, Some((Token::Ident("not"), _))) {
-                        self.stream.next()?;
-                        negated = true;
-                    }
+                    let negated =
+                        if matches!(self.stream.current()?, Some((Token::Ident("not"), _))) {
+                            self.stream.next()?;
+                            true
+                        } else {
+                            false
+                        };
                     let (name, span) =
                         expect_token!(self, Token::Ident(name) => name, "identifier")?;
                     let args = if matches!(self.stream.current()?, Some((Token::ParenOpen, _))) {
@@ -513,6 +520,10 @@ impl<'a> Parser<'a> {
                 self.parse_auto_escape()?,
                 self.stream.expand_span(span),
             ))),
+            Token::Ident("filter") => Ok(ast::Stmt::FilterBlock(Spanned::new(
+                self.parse_filter_block()?,
+                self.stream.expand_span(span),
+            ))),
             _ => syntax_error!("unknown block"),
         }
     }
@@ -690,6 +701,38 @@ impl<'a> Parser<'a> {
         Ok(ast::AutoEscape { enabled, body })
     }
 
+    fn parse_filter_block(&mut self) -> Result<ast::FilterBlock<'a>, Error> {
+        let mut filter = None;
+
+        while !matches!(self.stream.current()?, Some((Token::BlockEnd(..), _))) {
+            if filter.is_some() {
+                expect_token!(self, Token::Pipe, "`|`")?;
+            }
+            let (name, span) = expect_token!(self, Token::Ident(name) => name, "identifier")?;
+            let args = if matches!(self.stream.current()?, Some((Token::ParenOpen, _))) {
+                self.parse_args()?
+            } else {
+                Vec::new()
+            };
+            filter = Some(ast::Expr::Filter(Spanned::new(
+                ast::Filter {
+                    name,
+                    expr: filter,
+                    args,
+                },
+                self.stream.expand_span(span),
+            )));
+        }
+
+        let filter = filter
+            .ok_or_else(|| Error::new(ErrorKind::InvalidSyntax, "filter block without filter"))?;
+
+        expect_token!(self, Token::BlockEnd(..), "end of block")?;
+        let body = self.subparse(|tok| matches!(tok, Token::Ident("endfilter")))?;
+        self.stream.next()?;
+        Ok(ast::FilterBlock { filter, body })
+    }
+
     fn subparse<F: FnMut(&Token) -> bool>(
         &mut self,
         mut end_check: F,
@@ -739,11 +782,14 @@ impl<'a> Parser<'a> {
 }
 
 /// Parses a template
-pub fn parse<'a>(source: &'a str, filename: &'a str) -> Result<ast::Stmt<'a>, Error> {
-    let mut parser = Parser::new(source, filename, false);
+pub fn parse<'source, 'name>(
+    source: &'source str,
+    filename: &'name str,
+) -> Result<ast::Stmt<'source>, Error> {
+    let mut parser = Parser::new(source, false);
     parser.parse().map_err(|mut err| {
         if err.line().is_none() {
-            err.set_location(parser.filename, parser.stream.current_span().start_line)
+            err.set_location(filename, parser.stream.current_span().start_line)
         }
         err
     })
@@ -751,10 +797,10 @@ pub fn parse<'a>(source: &'a str, filename: &'a str) -> Result<ast::Stmt<'a>, Er
 
 /// Parses an expression
 pub fn parse_expr(source: &str) -> Result<ast::Expr<'_>, Error> {
-    let mut parser = Parser::new(source, "<expression>", true);
+    let mut parser = Parser::new(source, true);
     parser.parse_expr().map_err(|mut err| {
         if err.line().is_none() {
-            err.set_location(parser.filename, parser.stream.current_span().start_line)
+            err.set_location("<expression>", parser.stream.current_span().start_line)
         }
         err
     })
