@@ -5,6 +5,23 @@ use std::str::Chars;
 
 use crate::error::{Error, ErrorKind};
 
+// the type we want to use for refcounting
+#[cfg(feature = "sync")]
+pub(crate) type RcType<T> = std::sync::Arc<T>;
+#[cfg(not(feature = "sync"))]
+pub(crate) type RcType<T> = std::rc::Rc<T>;
+
+// we target Rust 1.41 and that does not have this macro yet
+macro_rules! _matches {
+    ($expression:expr, $( $pattern:pat )|+ $( if $guard: expr )? $(,)?) => {
+        match $expression {
+            $( $pattern )|+ $( if $guard )? => true,
+            _ => false
+        }
+    }
+}
+pub(crate) use _matches as matches;
+
 pub fn memchr(haystack: &[u8], needle: u8) -> Option<usize> {
     #[cfg(feature = "memchr")]
     {
@@ -43,36 +60,46 @@ pub struct HtmlEscape<'a>(pub &'a str);
 
 impl<'a> fmt::Display for HtmlEscape<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        #[cfg(feature = "v_htmlescape")]
+        {
+            fmt::Display::fmt(&v_htmlescape::escape(self.0), f)
+        }
         // this is taken from askama-escape
-        let bytes = self.0.as_bytes();
-        let mut start = 0;
+        #[cfg(not(feature = "v_htmlescape"))]
+        {
+            let bytes = self.0.as_bytes();
+            let mut start = 0;
 
-        for (i, b) in bytes.iter().enumerate() {
-            macro_rules! escaping_body {
-                ($quote:expr) => {{
-                    if start < i {
-                        f.write_str(unsafe { std::str::from_utf8_unchecked(&bytes[start..i]) })?;
+            for (i, b) in bytes.iter().enumerate() {
+                macro_rules! escaping_body {
+                    ($quote:expr) => {{
+                        if start < i {
+                            f.write_str(unsafe {
+                                std::str::from_utf8_unchecked(&bytes[start..i])
+                            })?;
+                        }
+                        f.write_str($quote)?;
+                        start = i + 1;
+                    }};
+                }
+                if b.wrapping_sub(b'"') <= b'>' - b'"' {
+                    match *b {
+                        b'<' => escaping_body!("&lt;"),
+                        b'>' => escaping_body!("&gt;"),
+                        b'&' => escaping_body!("&amp;"),
+                        b'"' => escaping_body!("&quot;"),
+                        b'\'' => escaping_body!("&#x27;"),
+                        b'/' => escaping_body!("&#x2f;"),
+                        _ => (),
                     }
-                    f.write_str($quote)?;
-                    start = i + 1;
-                }};
-            }
-            if b.wrapping_sub(b'"') <= b'>' - b'"' {
-                match *b {
-                    b'<' => escaping_body!("&lt;"),
-                    b'>' => escaping_body!("&gt;"),
-                    b'&' => escaping_body!("&amp;"),
-                    b'"' => escaping_body!("&quot;"),
-                    b'\'' => escaping_body!("&#x27;"),
-                    _ => (),
                 }
             }
-        }
 
-        if start < bytes.len() {
-            f.write_str(unsafe { std::str::from_utf8_unchecked(&bytes[start..]) })
-        } else {
-            Ok(())
+            if start < bytes.len() {
+                f.write_str(unsafe { std::str::from_utf8_unchecked(&bytes[start..]) })
+            } else {
+                Ok(())
+            }
         }
     }
 }
@@ -158,9 +185,9 @@ pub fn unescape(s: &str) -> Result<String, Error> {
 
 #[test]
 fn test_html_escape() {
-    let input = "<>&\"'";
+    let input = "<>&\"'/";
     let output = HtmlEscape(input).to_string();
-    assert_eq!(output, "&lt;&gt;&amp;&quot;&#x27;");
+    assert_eq!(output, "&lt;&gt;&amp;&quot;&#x27;&#x2f;");
 }
 
 #[test]
