@@ -9,7 +9,8 @@ use crate::environment::Environment;
 use crate::error::{Error, ErrorKind};
 use crate::instructions::{Instruction, Instructions};
 use crate::key::Key;
-use crate::value::{self, DynamicObject, Primitive, RcType, Value, ValueIterator};
+use crate::utils::{matches, RcType};
+use crate::value::{self, DynamicObject, Primitive, Value, ValueIterator};
 use crate::AutoEscape;
 
 #[derive(Debug)]
@@ -19,7 +20,7 @@ pub struct LoopState {
 }
 
 impl DynamicObject for LoopState {
-    fn fields(&self) -> &'static [&'static str] {
+    fn fields(&self) -> &[&str] {
         &[
             "index0",
             "index",
@@ -69,18 +70,18 @@ impl fmt::Display for LoopState {
 }
 
 #[derive(Debug)]
-pub struct Loop<'source> {
-    locals: BTreeMap<&'source str, Value>,
+pub struct Loop<'env> {
+    locals: BTreeMap<&'env str, Value>,
     with_loop_var: bool,
     iterator: ValueIterator,
     controller: RcType<LoopState>,
 }
 
 #[derive(Debug)]
-pub enum Frame<'source, 'context> {
+pub enum Frame<'env, 'context> {
     // This layer dispatches to another context
     Chained {
-        base: &'context Context<'source, 'context>,
+        base: &'context Context<'env, 'context>,
     },
     // this layer isolates
     Isolate {
@@ -91,7 +92,7 @@ pub enum Frame<'source, 'context> {
         value: Value,
     },
     // this layer is a for loop
-    Loop(Loop<'source>),
+    Loop(Loop<'env>),
 }
 
 #[derive(Debug, Default)]
@@ -118,13 +119,13 @@ impl Stack {
 }
 
 #[derive(Default, Debug)]
-pub struct Context<'source, 'context> {
-    stack: Vec<Frame<'source, 'context>>,
+pub struct Context<'env, 'context> {
+    stack: Vec<Frame<'env, 'context>>,
 }
 
-impl<'source, 'context> Context<'source, 'context> {
+impl<'env, 'context> Context<'env, 'context> {
     /// Stores a variable in the context.
-    pub fn store(&mut self, key: &'source str, value: Value) {
+    pub fn store(&mut self, key: &'env str, value: Value) {
         self.current_loop().locals.insert(key, value);
     }
 
@@ -166,7 +167,7 @@ impl<'source, 'context> Context<'source, 'context> {
     }
 
     /// Pushes a new layer.
-    pub fn push_frame(&mut self, layer: Frame<'source, 'context>) {
+    pub fn push_frame(&mut self, layer: Frame<'env, 'context>) {
         self.stack.push(layer);
     }
 
@@ -176,7 +177,7 @@ impl<'source, 'context> Context<'source, 'context> {
     }
 
     /// Returns the current innermost loop.
-    pub fn current_loop(&mut self) -> &mut Loop<'source> {
+    pub fn current_loop(&mut self) -> &mut Loop<'env> {
         self.stack
             .iter_mut()
             .rev()
@@ -191,24 +192,24 @@ impl<'source, 'context> Context<'source, 'context> {
 
 /// Helps to evaluate something.
 #[derive(Debug)]
-pub struct Vm<'env, 'source> {
-    env: &'env Environment<'source>,
+pub struct Vm<'env> {
+    env: &'env Environment<'env>,
 }
 
-impl<'env, 'source> Vm<'env, 'source> {
+impl<'env> Vm<'env> {
     /// Creates a new VM.
-    pub fn new(env: &'env Environment<'source>) -> Vm<'env, 'source> {
+    pub fn new(env: &'env Environment<'env>) -> Vm<'env> {
         Vm { env }
     }
 
     /// Evaluates the given inputs
-    pub fn eval<W: Write, S: Serialize>(
+    pub fn eval<S: Serialize>(
         &self,
-        instructions: &Instructions<'source>,
+        instructions: &Instructions<'env>,
         root: S,
-        blocks: &BTreeMap<&'source str, Instructions<'source>>,
+        blocks: &BTreeMap<&'env str, Instructions<'env>>,
         initial_auto_escape: AutoEscape,
-        output: &mut W,
+        output: &mut String,
     ) -> Result<Option<Value>, Error> {
         let mut context = Context::default();
         let root = Value::from_serializable(&root);
@@ -229,24 +230,21 @@ impl<'env, 'source> Vm<'env, 'source> {
     }
 
     /// This is the actual evaluation loop that works with a specific context.
-    fn eval_context<'context, W: Write>(
+    fn eval_context(
         &self,
-        mut instructions: &'env Instructions<'source>,
-        context: &'context mut Context<'source, 'context>,
-        blocks: &BTreeMap<&'source str, Vec<&'env Instructions<'source>>>,
-        block_stack: &mut Vec<&'source str>,
+        mut instructions: &'env Instructions<'env>,
+        context: &mut Context<'env, '_>,
+        blocks: &BTreeMap<&'env str, Vec<&'env Instructions<'env>>>,
+        block_stack: &mut Vec<&'env str>,
         initial_auto_escape: AutoEscape,
-        output: &mut W,
-    ) -> Result<Option<Value>, Error>
-    where
-        'source: 'context,
-        'env: 'context,
-    {
+        output: &mut String,
+    ) -> Result<Option<Value>, Error> {
         let mut pc = 0;
         let mut stack = Stack::default();
         let mut blocks = blocks.clone();
         let mut auto_escape = initial_auto_escape;
         let mut auto_escape_stack = vec![];
+        let mut capture_stack = vec![];
 
         macro_rules! bail {
             ($err:expr) => {{
@@ -283,6 +281,30 @@ impl<'env, 'source> Vm<'env, 'source> {
             }};
         }
 
+        macro_rules! out {
+            () => {
+                capture_stack.last_mut().unwrap_or(output)
+            };
+        }
+
+        macro_rules! begin_capture {
+            () => {
+                capture_stack.push(String::new());
+            };
+        }
+
+        macro_rules! end_capture {
+            () => {{
+                let captured = capture_stack.pop().unwrap();
+                // TODO: this should take the right auto escapine flag into account
+                stack.push(if !matches!(auto_escape, AutoEscape::None) {
+                    Value::from_safe_string(captured)
+                } else {
+                    Value::from(captured)
+                });
+            }};
+        }
+
         macro_rules! sub_eval {
             ($instructions:expr) => {{
                 sub_eval!($instructions, &blocks, block_stack, auto_escape);
@@ -297,7 +319,7 @@ impl<'env, 'source> Vm<'env, 'source> {
                     $blocks,
                     $block_stack,
                     $auto_escape,
-                    output,
+                    out!(),
                 )?;
             }};
         }
@@ -305,10 +327,10 @@ impl<'env, 'source> Vm<'env, 'source> {
         while let Some(instr) = instructions.get(pc) {
             match instr {
                 Instruction::EmitRaw(val) => {
-                    write!(output, "{}", val).unwrap();
+                    write!(out!(), "{}", val).unwrap();
                 }
                 Instruction::Emit => {
-                    try_ctx!(self.env.finalize(&stack.pop(), auto_escape, output));
+                    try_ctx!(self.env.finalize(&stack.pop(), auto_escape, out!()));
                 }
                 Instruction::StoreLocal(name) => {
                     context.store(name, stack.pop());
@@ -459,7 +481,10 @@ impl<'env, 'source> Vm<'env, 'source> {
                         let instructions = layers.first().unwrap();
                         sub_eval!(instructions);
                     } else {
-                        panic!("attempted to evaluate unreferenced block");
+                        bail!(Error::new(
+                            ErrorKind::ImpossibleOperation,
+                            "tried to invoke unknown block"
+                        ));
                     }
                     block_stack.pop();
                 }
@@ -467,10 +492,13 @@ impl<'env, 'source> Vm<'env, 'source> {
                     let name = stack.pop();
                     let tmpl = try_ctx!(name
                         .as_str()
-                        .and_then(|name| self.env.get_template(name))
                         .ok_or_else(|| {
-                            Error::new(ErrorKind::TemplateNotFound, "could not find template")
-                        }));
+                            Error::new(
+                                ErrorKind::ImpossibleOperation,
+                                "template name was not a string",
+                            )
+                        })
+                        .and_then(|name| self.env.get_template(name)));
 
                     // first load the blocks
                     for (name, instr) in tmpl.blocks().iter() {
@@ -489,10 +517,13 @@ impl<'env, 'source> Vm<'env, 'source> {
                     let name = stack.pop();
                     let tmpl = try_ctx!(name
                         .as_str()
-                        .and_then(|name| self.env.get_template(name))
                         .ok_or_else(|| {
-                            Error::new(ErrorKind::TemplateNotFound, "could not find template")
-                        }));
+                            Error::new(
+                                ErrorKind::ImpossibleOperation,
+                                "template name was not a string",
+                            )
+                        })
+                        .and_then(|name| self.env.get_template(name)));
                     let instructions = tmpl.instructions();
                     let mut referenced_blocks = BTreeMap::new();
                     for (&name, instr) in tmpl.blocks().iter() {
@@ -532,6 +563,12 @@ impl<'env, 'source> Vm<'env, 'source> {
                 Instruction::PopAutoEscape => {
                     auto_escape = auto_escape_stack.pop().unwrap();
                 }
+                Instruction::BeginCapture => {
+                    begin_capture!();
+                }
+                Instruction::EndCapture => {
+                    end_capture!();
+                }
                 Instruction::ApplyFilter(name) => {
                     let args = try_ctx!(stack.pop().try_into_vec());
                     let value = stack.pop();
@@ -554,7 +591,9 @@ impl<'env, 'source> Vm<'env, 'source> {
                         if let Some(layers) = inner_blocks.get_mut(name) {
                             layers.remove(0);
                             let instructions = layers.first().unwrap();
+                            begin_capture!();
                             sub_eval!(instructions);
+                            end_capture!();
                         } else {
                             panic!("attempted to super unreferenced block");
                         }
@@ -597,10 +636,10 @@ impl<'env, 'source> Vm<'env, 'source> {
 
 /// Simple version of eval without environment or vm.
 #[cfg(feature = "unstable_machinery")]
-pub fn simple_eval<W: Write, S: Serialize>(
+pub fn simple_eval<S: Serialize>(
     instructions: &Instructions<'_>,
     root: S,
-    output: &mut W,
+    output: &mut String,
 ) -> Result<Option<Value>, Error> {
     let env = Environment::new();
     let empty_blocks = BTreeMap::new();
